import 'package:json_annotation/json_annotation.dart';
import 'dart:math';

part 'codenames_board_base.g.dart';

@JsonSerializable()
class Board {
  // ROW MAJOR
  List<Tile> board;
  String turn;
  int w, h;
  String startingTeam;

  bool ended;

  void click(int x, int y) {
    Tile t = board[y * w + x];
    if (t.revealed) return;
    t.revealed = true;

    if (t.team == Team.BLACK) {
      ended = true;
      return;
    }

    if (t.team != turn) swap(); // wrong guess
  }

  void swap() {
    turn = Team.opposite(turn);
  }

  factory Board.filled(int w, int h, List<String> possibleWords, Random r) {
    Board b = Board();

    b.startingTeam = choose<String>([Team.RED, Team.BLUE], r);
    b.ended = false;

    List<String> used = [];

    int amt = w * h;

    b.board = List.generate(amt, (_) {
      String choice = choose<String>(possibleWords, r, used);
      used.add(choice);
      return Tile(choice);
    });

    List<String> assignments = [];
    assignments.addAll(List.filled(7, Team.RED));
    assignments.addAll(List.filled(7, Team.BLUE));
    assignments.addAll(List.filled(1, Team.BLACK));
    assignments.addAll(List.filled(1, b.startingTeam));
    assignments.addAll(List.filled(amt - assignments.length, Team.BLACK));

    assignments.shuffle(r);

    for (int i = 0; i < assignments.length; i++) {
      b.board[i].team = assignments[i];
    }

    return b;
  }

  Board();

  factory Board.fromJson(Map<String, dynamic> json) => _$BoardFromJson(json);

  Map<String, dynamic> toJson() => _$BoardToJson(this);
}

@JsonSerializable()
class Tile {
  String team;
  bool revealed;
  String content;

  Tile([String this.content]) : revealed = false;

  factory Tile.fromJson(Map<String, dynamic> json) => _$TileFromJson(json);

  Map<String, dynamic> toJson() => _$TileToJson(this);
}

@JsonSerializable()
class GameRequest {
  String type;
  String gameID;
  Map<String, dynamic> data;

  GameRequest([this.type, this.data, String gameID]);

  factory GameRequest.fromJson(Map<String, dynamic> json) =>
      _$GameRequestFromJson(json);

  Map<String, dynamic> toJson() => _$GameRequestToJson(this);
}

/**
 * type: GET_BOARD, data: {}
 * type: PICK, data: {point: [x, y]}    
 * type: SWITCH_TEAM, data: {team: "team"}
 * type: RESET, data: {}
 */
class RequestType {
  static const String GET_BOARD = "GET_BOARD",
      PICK = "PICK",
      SWITCH_TEAM = "SWITCH_TEAM",
      RESET = "RESET";
}

class Team {
  static const String RED = "RED",
      BLUE = "BLUE",
      NEUTRAL = "NEUTRAL",
      BLACK = "DEATH";

  static String opposite(String team) {
    switch (team) {
      case BLUE:
        return RED;
      case RED:
        return BLUE;
      default:
        return team;
    }
  }
}

T choose<T>(List<T> list, Random r, [List<T> exclude = const []]) {
  T choice;
  while (exclude.contains(choice = list[r.nextInt(list.length)]));
  return choice;
}
